#include "BPatch.h"
#include "BPatch_addressSpace.h"
#include "BPatch_process.h"
#include "BPatch_binaryEdit.h"
#include "BPatch_function.h"
#include "BPatch_point.h"
#include "BPatch_flowGraph.h"

#include <string>

// Example 1: create an instance of class BPatch
BPatch bpatch;

// Example 2: attaching, creating, or opening a file for rewrite
typedef enum {
    create,
    attach,
    open
} accessType_t;

BPatch_addressSpace *startInstrumenting(accessType_t accessType,
                                        const char *name,
                                        int         pid,    // For attach
                                        const char *argv[]) // For create
{
    BPatch_addressSpace *handle = NULL;
    switch (accessType) {
    case create:
        handle = bpatch.processCreate(name, argv);
        break;
    case attach:
        handle = bpatch.processAttach(name, pid);
        break;
    case open:
        handle = bpatch.openBinary(name);
        break;
    }
    return handle;
}


// Example 2: find the entry point for "foo"
std::vector<BPatch_point *> *findEntryPoint(BPatch_addressSpace *app)
{
    std::vector<BPatch_function *> functions;
    std::vector<BPatch_point *> *points;

    BPatch_image *appImage = app->getImage();
    appImage->findFunction("foo", functions, true, true, true);
    points = functions[0]->findPoint(BPatch_entry);

    return points;
}

BPatch_variableExpr *intCounter;
// Example 3: create and insert an increment snippet
void createAndInsertSnippet(BPatch_addressSpace *app,
                            std::vector<BPatch_point *> *points)
{
    BPatch_image *appImage = app->getImage();
    intCounter = app->malloc(*(appImage->findType("int")));
    BPatch_arithExpr addOne(BPatch_assign, *intCounter,
                            BPatch_arithExpr(BPatch_plus, *intCounter,
                                             BPatch_constExpr(1)));

    app->insertSnippet(addOne, *points);
}


void printCount(BPatch_thread *thread, BPatch_exitType)
{
    (void)thread;
    int val = 0;
    intCounter->readValue(&val, sizeof(int));
    fprintf(stderr, "Function executed %d times.\n", val);
}


// Example 4: finish things up (continue or write out)
void finishInstrumenting(BPatch_addressSpace *app, const char *newName)
{
    BPatch_process    *appProc = dynamic_cast<BPatch_process *>(app);
    BPatch_binaryEdit *appBin  = dynamic_cast<BPatch_binaryEdit *>(app);
    if (appProc) {
        appProc->continueExecution();
        while (!appProc->isTerminated()) {
            bpatch.waitForStatusChange();
        }
    }
    if (appBin) {
        appBin->writeFile(newName);
    }
}



int main(void)
{
    const char *progName = "hello"; // = ...
    int progPID = 42; // = ...
    const char *progArgv[] = {"hello", NULL}; // = ...
    // Example 1: create/attach/open a binary
    BPatch_addressSpace *app = startInstrumenting(create, // or attach or open
                                                  progName,
                                                  progPID,
                                                  progArgv);

    // Example 2: get entry point
    std::vector<BPatch_point *> *entryPoint = findEntryPoint(app);

    // Example 3: create and insert increment snippet
    createAndInsertSnippet(app, entryPoint);

    bpatch.registerExitCallback(printCount);

    // Example 4: finish up instrumenting
    finishInstrumenting(app, progName);

    // Example 5: get a count of memory accesses
    //int insns_access_memory = binaryAnalysis(app);

    // Example 6: instrument memory accesses
    //instrumentMemory(app);

    return 0;
}
